/*-----------------------------------------------------------------------------

 2002, Jeff M Soft und Hardware GmbH, All Rights Reserved

-----------------------------------------------------------------------------*/

#include "Module.h"
#include "SEMod_struct.h"
#include "SEPin.h"
#include "GuiModule.h"
#include <assert.h>
#include <stdio.h>

#define TABLE_SIZE 512

// define some constants to make referencing in/outs clearer
#define PN_IN1 1
#define PN_OUT1 2
#define PN_PATCH_PARAM 0

void Module::OnPlugStateChange(SEPin *pin)
{
	if( pin->getPinID() == PN_PATCH_PARAM )
	{
		FillLookupTable();
		OnLookupTableChanged();
	}

	if( pin->getPinID() == PN_IN1 )
	{
		ChooseProcess();
		getPin(PN_OUT1)->TransmitStatusChange( SampleClock(), getPin(PN_IN1)->getStatus() );
	}
}

void Module::ChooseProcess()
{
	if( getPin(PN_IN1)->getStatus() == ST_RUN )
	{
		SET_PROCESS_FUNC(Module::sub_process);
	}
	else
	{
		static_count = getBlockSize();
		SET_PROCESS_FUNC(Module::sub_process_static);
	}
}

void Module::open()
{
	SetupLookupTable();

	SET_PROCESS_FUNC(Module::sub_process);

	SEModule_base::open();	// always call the base class

/* testing
	// calulate how much memory needed.
	int memorySize = (sizeof(myArrayType) + 3) / 4;

	// 'allocate' the required memory.
	CallHost (seaudioMasterCreateSharedLookup, (long) &myArrayRawPointer, memorySize, "myArray", getSampleRate());

	// cast pointer to more friendly type.
	myArrayType& MyArray = *((myArrayType*) myArrayRawPointer);

	// now use it like any regular array.
	MyArray[3][4] = 42.0f;
*/
}


void Module::sub_process(long start_pos, long sampleframes)
{
	float *in1 = in1_ptr + start_pos;
	float *out1 = out1_ptr + start_pos;

	const float half = 0.5f;
	const float l_table_size = 512.f;
	int int_idx;

	short control_word;
	short control_word2;
	__asm
	{
		fstcw	control_word		// store fpu control word 
		mov 	dx, word ptr [control_word]
 		or		dx,0x0400			// round towards -oo 
		mov		control_word2, dx
		fldcw	control_word2		// load modfied control word 
	}

	for( int s = sampleframes ; s > 0 ; s-- )
	{
		/*
		FSAMPLE count = *in++ + 0.5f; // map +/-5 volt to 0.0 -> 1.0
		count = max( count , 0.f );
		count = min( count, 1.0f );

		float index = count * 512.f;
		float idx_frac = index - floorf(index);
		float *p = l_wave_data + (int)index; // keep top 9 bits as index into 512 entry wavetable
		//float s 1 = *p++;
		//*out++ = s1 * (1.f - idx_frac) + *p * idx_frac;
		*out++ = p[0] + (p[1] - p[0]) * ((*(float*)&itemp) - 1.f);
		*/
		__asm
		{
			mov     eax,dword ptr [in1] 
			// multiply index by table size
			fld		dword ptr [eax]
			fadd	dword ptr [half]
			fmul	dword ptr l_table_size
			mov   ecx, dword ptr [this]

			fist   dword ptr [int_idx]	// int_idx = (int) idx
			fisub   dword ptr [int_idx]	// frac = idx - int_idx

			mov	  ecx, dword ptr[ecx]this.m_lookup_table
			mov	  edx, dword ptr [int_idx]

			cmp		edx, 0
			jge		not_too_small
			mov     edx, 0
			fstp	st(0)
			fldz
			jmp		in_range
not_too_small:
			cmp		edx, 512
			jl		in_range
			mov     edx, 512
			fstp	st(0)
			fldz
in_range:
			// perform table lookup, interpolate
			lea		esi,[ecx+edx*4]		// esi = & (table[int_idx] )
			fld		dword ptr [esi]     // table[0] 
			fld		dword ptr [esi+4]   // table[1] 
			fsub	st(0),st(1)			// table[1] - table[0]
			mov         ecx,dword ptr [out1] 
			fmulp	st(2),st(0)			// * idx_frac 
			fadd						// + table[0]

			// inc pointers, store result
			add         eax,4 
			mov         dword ptr [in1],eax 
			fstp        dword ptr [ecx] 

			add         ecx,4
			mov         dword ptr [out1],ecx 
		}
	}

	// restore original control word 
	__asm
	{
		fldcw	control_word
	}
}

void Module::OnLookupTableChanged()
{
//	_RPT1(_CRT_WARN, "OnLookupTableChanged() %x\n",this );

	ChooseProcess();
	if( getPin(PN_IN1)->getStatus() != ST_RUN )
	{
		// can't add event if sleeping
//		OutputChange( SampleClock(), GetPlug( PN_OUT1 ), ST_ONE_OFF );
		getPin(PN_OUT1)->TransmitStatusChange( SampleClock(), ST_ONE_OFF );
	}
}

void Module::sub_process_static(long start_pos, long sampleframes)
{
	sub_process( start_pos, sampleframes );

//	SleepIfOutputStatic(sampleframes);
	static_count = static_count - sampleframes;
	if( static_count <= 0 )
	{
		CallHost(seaudioMasterSleepMode);
	}
}

/*
 If you use several waveshapers, each may be set to a different curve, so each needs it's own lookup table.  You need to give each table a unique name..

e.g. "Waveshaper 1 curve", "Waveshaper 2 curve", etc

SE assigns every module a unique number (handle).  If your module has a GUI object and a DSP object, they share the same handle.
This is how SE know where to send GUI->Module communication.

Each waveshaper needs a unique number to identify it's wavetable.  The handle is ideal for this purpose.
*/
int Module::Handle(void)
{
	return CallHost(seaudioMasterGetModuleHandle);
}

void Module::SetupLookupTable()
{
	char table_name[40];

	sprintf(table_name, "SE wave shaper %x curve", Handle() );

	bool need_initialise = CreateSharedLookup( table_name, &m_lookup_table, -1, TABLE_SIZE+2 );
}

bool Module::CreateSharedLookup( char *p_table_name, void *p_table_pointer, float p_sample_rate, int p_size)
{
	return 0 != CallHost(seaudioMasterCreateSharedLookup, (long)p_table_pointer, p_size, p_table_name, p_sample_rate);
}

void Module::FillLookupTable()
{
//	mypoint2 nodes[WS_NODE_COUNT]; // x,y co-ords of control points
	mypoint nodes[WS_NODE_COUNT]; // x,y co-ords of control points
	GuiModule::UpdateNodes( nodes, SeSdkString(shape_ptr) );

/* Old slower code
	int segments = 11;
	int from = 0; // first x co=ord
	for( int i = 1 ; i < segments; i++ )
	{
		float to = nodes[i].x * TABLE_SIZE *0.01f; // convert to table index (0-512)
		
		if( to >= TABLE_SIZE )
			to = TABLE_SIZE - 1;

		for( int t = from ; t < to ; t++ )
		{
			float gain = nodes[i-1].y + ((t-from)/(to-from) )*(nodes[i].y - nodes[i-1].y);
			gain = ( 50.f - gain ) / 100.f;
//			m_lookup->SetValue( t, (FSAMPLE)gain );
			m_lookup_table[t] = gain;
//			_RPT2(_CRT_WARN, "%d,%d\n", (int)t,(int)gain );
		}

		from = (int) to;
	}
*/

	// new, faster code
	int segments = 11;
	double from_x = 0.f; // first x co=ord
	int table_index = 0;
	for( int i = 1 ; i < segments; ++i )
	{
		double to_x = nodes[i].x * TABLE_SIZE * 0.01;
		double delta_y = nodes[i].y - nodes[i-1].y;
		double delta_x = from_x - to_x;
		double slope = 0.01 * delta_y / delta_x; 
		double c = 0.5 - 0.01 * nodes[i-1].y - from_x * slope; 

		int int_to = (int) to_x;

		if( int_to > TABLE_SIZE )
			int_to = TABLE_SIZE;

		if( i == segments - 1 )
			int_to = TABLE_SIZE + 1; // fill in last 'extra' sample.

		for( ; table_index < int_to ; table_index++ )
		{
			m_lookup_table[table_index] = (float) (( (double)table_index ) * slope + c);
//			_RPT2(_CRT_WARN, "%3d : %f\n", table_index, m_lookup_table[table_index] );
		}

		from_x = to_x;
	}
}

Module::Module(seaudioMasterCallback2 seaudioMaster, void *p_resvd1) : SEModule_base(seaudioMaster, p_resvd1)
{
}

// describe the pins (plugs and parameters)
bool Module::getPinProperties (long index, SEPinProperties* properties)
{
	/*
	LIST_PIN2("Signal In", in1_ptr, DR_IN, "0","", IO_POLYPHONIC_ACTIVE,"");
	LIST_PIN2("Signal Out", out1_ptr, DR_OUT, "","", 0,"");
	// move this to patch mgr
	LIST_VAR3("Ignore Program Change", ignore_prog_change, DR_PARAMETER, DT_BOOL ,"0","", 0,"Prevents this control responding to patch changes from 'Patch Select' module");
	LIST_VAR3("shape", m_shape_string,  DR_IN, DT_TEXT ,"", "", IO_UI_COMMUNICATION/*| IO_PRIVATE* /,"");

	*/
	switch( index )								// !!TODO!! list your in / out plugs
	{
	case 0:
		properties->name				= "shape";
		properties->variable_address	= &shape_ptr;
		properties->direction			= DR_IN;
		properties->datatype			= DT_TEXT;
//		properties->default_value		= "";
		properties->flags				= IO_UI_COMMUNICATION_DUAL| IO_PRIVATE;
		break;
	case 1:
		properties->name				= "Signal In";
		properties->variable_address	= &in1_ptr;
		properties->direction			= DR_IN;
		properties->datatype			= DT_FSAMPLE;
		properties->flags				= IO_POLYPHONIC_ACTIVE;
		break;
	case 2:
		properties->name				= "Signal Out";
		properties->variable_address	= &out1_ptr;
		properties->direction			= DR_OUT;
		properties->datatype			= DT_FSAMPLE;
		break;
	default:
		return false; // host will ask for plugs 0,1,2,3 etc. return false to signal when done
	};

	return true;
}
